//*****************************************************************************
// ComIO: Classe responsvel pelas funes de entrada e saida da interface serial 
//
// Argumentos:
//
// Autor: Gatti - 23/11/2010 - v 1.0.0
//	- criao do modulo
//
// Autor: Gatti - 13/01/2012 - v 1.0.1
//	- converso para Visual Studio 2010
//
// Autor: Gatti - 27/11/2018 - v 1.1.0
//	- Migrao para Visual Studio 2017
//
// Autor: Gatti - 17/02/2020 - v 2.0.0
//	- Migrao para Visual Studio 2019
//	- Adaptao para nova verso da DLL simples
//
// Controle de Alteraes:
// <autor> - <data> - v <x.y>
//	- <descrio>
//
// Copyright  Infiza/Telemed 2010-2020
// todos os diretors reservados.
// proibida a utilizao/reproduo sem prvia autorizao.
//*****************************************************************************
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.IO;
using System.IO.Ports;
using System.Xml;
using System.Xml.Schema;
using System.Threading;
using System.Runtime.InteropServices;

namespace Telemed.Communication.Components {

	public class ComIO: SerialPort {

		private string K_Name = "ComIO";
		private string K_Version = "2.0.0";
		private string K_NameVersion = "";

		public static byte K_STX = 0x02;        // STX
		public static byte K_SOH = 0x01;        // SOH
		public static byte K_ETX = 0x03;        // ETX
		public static byte K_ACK = 0x06;        // ACK
		public static byte K_NCK = 0x15;        // NCK
		public static byte K_CR = 0x0D;         // CR

		private string port = "COM1";
		private int baudRate = 9600;
		private Parity parity = Parity.None;
		private bool dtrControl = true;
		private bool rtsEnable = true;
		private Handshake handshake = Handshake.None;
		private StopBits stopBits = StopBits.One;
		private int dataBits = 8;
		private int maxBuffer = 100;

		private byte[] receiveBuf;

		public ComIO(string pPort) {
			K_Name = K_Name + "." + pPort;
			K_NameVersion = K_Name + "-[" + K_Version + "]";

			port = pPort;
			PortName = pPort;
			BaudRate = baudRate;
			DtrEnable = dtrControl;
			Handshake = handshake;
			RtsEnable = rtsEnable;
			Parity = parity;
			StopBits = stopBits;
			DataBits = dataBits;

			receiveBuf = new byte[maxBuffer];

			Console.WriteLine(string.Format("{0}.Construct - Port: '{1}'", K_NameVersion, port));
		}

		public new void Open() {
			if(!IsOpen) {
				try {
					base.Open();
					Console.WriteLine(string.Format("{0}.Open", K_Name));
				} catch(Exception ex) {
					base.Close();
					throw new Exception(string.Format("{0}.Open - Erro ao abrir canal serial", K_Name), ex); 
				}
			}
		}

		public new void Close() {
			if(IsOpen) {
				try {
					base.Close();
					Console.WriteLine(string.Format("{0}.Close", K_Name));
				} catch(Exception ex) {
					throw new Exception(string.Format("{0}.Close - Erro ao fechar canal serial", K_Name), ex);
				}
			}
		}

		public void Send(byte[] pMessage, int pTimeout) {
			byte ack;

			try {
				Console.WriteLine(K_Name + ".TX: " + MessageToString(pMessage));

				clear();				// Limpa os buffers da porta serial (input e output)
				SetTimeout(pTimeout);	// Inicia Temporizao para envio

				// Envia o Buffer - byte a byte com delay entre eles para nao sobrecarregar o equipamento
				for(int i = 0; i < pMessage.Length; i++) {
					Write(pMessage, i, 1);
					Thread.Sleep(300);
				}

				// Espera a resposta do equipamento (ack ou nack)
				ack = (byte)ReadByte();
				if(ack == Message.K_NCK) {          // se receber nack - ento descarta a mensagem e avisa a aplicao
					string s = K_Name + ".TX: Recebido - <nck>";
					Console.WriteLine(s);
					throw new Exception(s);
				} else if(ack != Message.K_ACK){	// se receber algo diferente de ack ou nack - entao descata a mensagem e avisa a aplicao
					clear();
					string s = string.Format(K_Name + ".TX: " + "Esperando <ack> ou <nck> - recebido: {0}", ack.ToString());
					Console.WriteLine(s);
					throw new Exception(s);
				}
			} catch(TimeoutException ex) {
				string s = K_Name + ".TX: Timeout";
				Console.WriteLine(s + " - " + ex.Message);		// equipamento nao respondeu nada - avisa a aplicao
				throw new Exception(s, ex);
			
			} catch(IOException ex) {               // algum erro de IO - avisa a aplicao
				string s = K_Name + ".TX: IOException";
				Console.WriteLine(s + " - " + ex.Message);
				throw new Exception(s, ex);
			
			} catch(Exception ex) {                 // qualquer outro erro - avisa a aplicao
				string s = K_Name + ".TX: Exception";
				Console.WriteLine(s + " - " + ex.Message);
				throw new Exception(s, ex);
			
			} finally {
				ResetTimeout();			// Finaliza as temporizaes
			}
		}

		public Byte[] Receive(int pTimeout) {
			byte lrc = 0;
			byte c = 0;
			int length = 0;

			try {
				// Recebe o <STX>, despresando qualquer outro valor
				SetTimeout(pTimeout);   // Inicia Temporizao para recebimento
				while((c = (byte)ReadByte()) != Message.K_STX) {
					Console.WriteLine(c);
				}

				// Recebe o campo <informao> at o <ETX> (exnclusive)
				SetTimeout(pTimeout);   // Reinicia 0 Temporizao para recebimento
				receiveBuf[length++] = c;
				while((c = (byte)ReadByte()) != Message.K_ETX) {
					if(length < (maxBuffer - 3)) {
						receiveBuf[length++] = c;
						lrc ^= c;
					}
				}
				// Recebe o <ETX>
				receiveBuf[length++] = c;
				lrc ^= c;

				// Recebe o <LRC>
				c = (byte)ReadByte();
				receiveBuf[length++] = c;

				if(lrc != c) {          // Se <LRC> calculado for diferente ao recebido ento mesangem com erro - avisa a aplicao
					Write(Message.K_NCK.ToString());    // Envia <NCK> para o equipamento

					string s = K_Name + ".RX: Transmitido - <nck>";
					Console.WriteLine(s);
					throw new Exception(s + string.Format(": Recebido: {0}, Calculado: {1}", c.ToString("X2"), lrc.ToString("X2")));
				}
			} catch(TimeoutException ex) {
				string s = K_Name + ".RX: Timeout";
				Console.WriteLine(s + " - " + ex.Message);       // equipamento nao respondeu nada - avisa a aplicao
				throw new Exception(s, ex);
			} catch(IOException ex) {                 // algum erro de IO - avisa a aplicao
				string s = K_Name + ".RX: IOException";
				Console.WriteLine(s + " - " + ex.Message);
				throw new Exception(s, ex);
			} catch(Exception ex) {                   // qualquer outro erro - avisa a aplicao
				string s = K_Name + ".RX: Exception";
				Console.WriteLine(s + " - " + ex.Message);
				throw new Exception(s, ex);
			} finally {           // Finaliza as temporizaes
				ResetTimeout();
			}

			// Copia a mensagem recebida para um novo buf
			byte[] buf = new byte[length];
			for(int j = 0; j < length; j++) { buf[j] = receiveBuf[j]; }

			Console.WriteLine(K_Name + ".RX: " + MessageToString(buf));

			return buf;
		}

		private void SetTimeout(int pTimeOut) {
			if(pTimeOut > 0) {
				WriteTimeout = pTimeOut;
				ReadTimeout = pTimeOut;
			} else {
				ResetTimeout();
			}
		}

		private void ResetTimeout() {
			WriteTimeout = SerialPort.InfiniteTimeout;
			ReadTimeout = SerialPort.InfiniteTimeout;
		}

		private void clear() {
			DiscardOutBuffer();
			DiscardInBuffer();
		}

		#region Rotina para Depurao
		private char ByteToChar(byte pByte) {
			return (char)pByte;
		}
		private string ByteToStringHexa(byte pByte) {
			return string.Format("<0x{0}>", pByte.ToString("X2"));
		}
		private string MessageToString(byte[] pBuf) {
			string s = "";

			int i;
			int bufLength = pBuf.Length;	// Tamanho total do buffer.

			i = 0;  // Posio do <STX> no buffer.
			s += pBuf[i] == K_STX ? "<stx>" : ByteToStringHexa(pBuf[i]);

			i = 1;  // Posio da <informao> no buffer.
			for(; i < bufLength - 2; i++) { s += ByteToChar(pBuf[i]);	}

			i = bufLength - 2;	// Posio do <ETX> no buffer.
			s += pBuf[i] == K_ETX ? "<etx>" : ByteToStringHexa(pBuf[i]);

			i = bufLength - 1;		// Posio do <LRC> no buffer.
			s += ByteToStringHexa(pBuf[i]);

			s += ((pBuf[0] == K_STX) && (pBuf[bufLength-2] == K_ETX)) ? "" : " - Mensagem com formato invlido.";

			return s;
		}
		#endregion  Rotina para Depurao
	}
}
